from utils import copy, saves_args
from resource import Resource, MarkerResource

class TaskException(Exception):
	pass


class Task(Resource):

	inputs = None
	outputs = None

	@saves_args
	def __init__(self, **kwargs):
		
		# update any class-level parameters with those provided in constructor
		self.parameters = kwargs

		# make a hashable signature of the parameters.  
		# This uniquely identifies the task
		self._hashable_parameters = tuple(sorted(self.parameters.items()))

		# Ensure all parameters are hashable
		try:
			if not all(
				[isinstance(hash(v), int) for k,v in self._hashable_parameters]
			):
				raise ValueError('values used in parameters must be hashable.')

		except TypeError:
			raise ValueError('values used in parameters must be hashable.')

		super(Task, self).__init__(**kwargs)



	def get_ready(self, lot, pilot, name, clobber=False):

		super(Task,self).get_ready(lot, pilot, name, clobber)

		# Ready the inputs
		self.inputs = copy(self._inputs())
		for input in self.get_all_inputs():
			input.get_ready(
				self.get_lot(), self.get_pilot(), 'poop', self.get_clobber())

		# Ready the outputs 
		self.outputs = copy(self._outputs())
		for output in self.get_all_outputs():
			output.get_ready(
				self.get_lot(), self.get_pilot(), 'poop', self.get_clobber())


	def get_all_inputs(self):

		if self.inputs is None:
			return []

		if isinstance(self.inputs, Resource):
			inputs = [self.inputs]

		elif isinstance(self.inputs, dict):
			inputs = self.inputs.values()

		else:
			inputs = self.inputs

		if not isinstance(inputs, (list, tuple)):
			raise TaskException(
				'inputs must be a Resource, or a list, tuple, or dict of '
				'resources.'
			)

		return inputs


	def get_all_outputs(self):

		if self.outputs is None:
			return []

		if isinstance(self.outputs, Resource):
			return [self.outputs]
			
		if isinstance(self.outputs, dict):
			return self.outputs.values()

		if isinstance(self.outputs, tuple):
			return list(self.outputs)

		if isinstance(self.outputs, list):
			return self.outputs

		raise TaskException(
			'outputs must be a Resource, or a list, tuple, or dict of '
			'resources.'
		)


	def exists(self):
		'''
			Indicates whether this Task is complete.  It's considered
			complete if all of its outputs `exist`.
		'''

		return all([o.exists() for o in self.get_all_outputs()])


	def _outputs(self):
		return self.outputs


	def _inputs(self):
		return self.inputs


	def _run(self, lot=None, pilot=False):
		return_val = self.run()
		self._after()
		

	def run(self):
		raise NotImplementedError('You need to define run().')


	def _after(self):
		pass


	def __hash__(self):
		return hash((self.__class__.__name__, self._hashable_parameters))



class MarkedTask(Task):

	def get_ready(self, lot, pilot, name, clobber=False):
		super(MarkedTask, self).get_ready(lot, pilot, name, clobber)

		try:
			path = self.marker_path
		except AttributeError:
			path = '.'

		fname = self.name + '.marker'
		self.marker = MarkerResource(path, fname)
		self.marker.get_ready(lot, pilot, name, clobber)


	def exists(self):
		return self.marker.exists()


	def _after(self):
		super(MarkedTask, self)._after()
		self.marker.mark()


class SimpleTask(MarkedTask):
	outputs = None
	marker_path = './linguini_markers'



import inspect
from functools import wraps


def copy(obj):
	if isinstance(obj, dict):
		new_obj = dict([(k, v.copy()) for k,v in obj.items()])

	elif isinstance(obj, (tuple, list)):
		new_obj = obj.__class__([v.copy() for v in obj])

	elif obj is None:
		new_obj = obj

	else:
		new_obj = obj.copy()

	return new_obj


def saves_args(f):
	@wraps(f)
	def wrapped(self, *args, **kwargs):
		if not hasattr(self, 'args'):
			argspec = inspect.getargspec(f)
			_args = locals()['args']
			_kwargs = locals()['kwargs']
			self.args = {'args': _args, 'kwargs': _kwargs}

		return f(self, *args, **kwargs)

	return wrapped









import os
from datetime import datetime
from utils import saves_args


class ResourceException(Exception):
	pass


class Resource(object):


	@saves_args
	def __init__(self, **kwargs):
		self.resolve_static(**kwargs)
		self.resolve_inherited(**kwargs)


	def resolve_inherited(self, **kwargs):
		self.instance_lot = kwargs.pop('lot', None)
		self.instance_pilot = kwargs.pop('pilot', None)
		self.instance_clobber = kwargs.pop('clobber', None)


	def resolve_static(self, **kwargs):
		# Resolve the `static`, `share`, and `ignore_pilot` attributes
		self.static = kwargs.pop('static', False)
		self.share = kwargs.pop('share', False)
		self.ignore_pilot = kwargs.pop('ignore_pilot', False)


	def get_ready(self, lot, pilot, name, clobber):

		# resolve propogation of the run properties
		self.inherited_lot = lot
		self.inherited_pilot = pilot
		self.inherited_clobber = clobber
		self.name = name

		# make sure that the lot is string-like or None
		lot_is_not_string = not isinstance(self.get_lot(), basestring)
		lot_is_not_none = self.get_lot() is not None
		if lot_is_not_string and lot_is_not_none:
			raise ValueError('`lot` must be string-like.')

		self._is_ready = True


	def is_ready(self):
		try:
			return self._is_ready
		except AttributeError:
			return False


	def get_clobber(self):
		if self.instance_clobber is not None:
			return self.instance_clobber

		if hasattr(self, 'clobber') and self.clobber is not None:
			return self.clobber

		if self.inherited_clobber is not None:
			return self.inherited_clobber

		raise ResourceException(
			'could not resolve clobber in %s' % self.__class__.__name__)


	def get_pilot(self):
		if self.ignore_pilot or self.static:
			return False

		if self.instance_pilot is not None:
			return self.instance_pilot

		if hasattr(self, 'pilot') and self.pilot is not None:
			return self.pilot

		if self.inherited_pilot is not None:
			return self.inherited_pilot

		raise ResourceException(
			'could not resolve pilot in %s' % self.__class__.__name__)


	def get_lot(self):
		if self.share or self.static:
			return None

		if self.instance_lot is not None:
			return self.instance_lot

		if hasattr(self, 'lot') and self.lot is not None:
			return self.lot

		if self.inherited_lot is not None:
			return self.inherited_lot

		return None



	def exists(self):
		raise NotImplementedError(
			'Resources need to implement an `exists` method.'
		)


	def copy(self):
		return self.__class__(*self.args['args'], **self.args['kwargs'])


class File(Resource):

	@saves_args
	def __init__(self, path, fname, **kwargs):
		self.path = path
		self.fname = fname
		super(File, self).__init__(**kwargs)


	def get_path(self):

		# ensure the resource is ready
		if not self.is_ready():
			raise ResourceException('Resource is not ready.')

		fname = (
			(('%s_' % self.get_lot()) if self.get_lot() is not None else '')
			+ ('pilot_' if self.get_pilot() else '')
			+ self.fname
		)
		return os.path.join(self.path, fname)


	def get_dir(self):
		return os.path.dirname(self.get_path())


	def exists(self):
		return os.path.isfile(self.get_path())

	def open(self, flags):

		# avoid clobbering files (if that's not what we're told to do)
		if 'w' in flags:

			# if writing, check whether the file exists
			if os.path.isfile(self.get_path()):
				if self.get_clobber():
					print '\t INFO: clobbered %s' % self.get_path()
				else:
					raise IOError(
						'File: by default, I refuse to overwrite files. '
						+ self.get_path()
					)

		# if we're opining in a write mode, then make dirs if needed
		if 'a' in flags or 'w' in flags:
			if not os.path.isdir(self.get_dir()):
				os.makedirs(self.get_dir())

		# hands over a plain file handle
		return open(self.get_path(), flags)


class MarkerResource(File):

	def mark(self):

		# if the marker path already exists, perfect
		if os.path.isdir(self.path):
			pass

		# if the marker path points to a file, not a dir, fail
		elif os.path.exists(self.path):
			raise IOError(
				('MarkerResource: I could not mark task `%s` as done because '
				'there is a file in the place of the directory I am supposed '
				'to write to') % self.name
			)

		# if the path doesn't exist yet, make it
		else:
			os.makedirs(self.path)

		self.open('a').write('%s\n' % str(datetime.now()))


class IncrementalFile(object):
	'''
		This is like a file resource, but we don't rely on the existence of
		the file to determine whether the resource is satisfied.  We use a
		separate "marker" file, that is created when the resource is marked
		as done (by calling mark_as_done()).  This file contains a timestamp
		of when the resource was marked done according to local machine's time
	'''

	@saves_args
	def __init__(self, path, fname, part_name):
		self.path = path
		self.fname = fname
		self.part_name = part_name

		
	def exists(self):
		return os.path.isfile(get_marker_path)


	def get_marker_path(self):
		return self.get_path() + '.' + self.part_name + '.marker'


	def mark_as_done(self):
		# note that this will create a marker file with the current time
		# stamp.  In general, it should never be called if there is an existing
		# file already by the same name, however, if that does happen somehow
		# it will append to the file.
		marker_fh = open(self.get_marker_path(), 'a')
		marker_fh.write('%s\n' % str(datetime.now()))




class Folder(File):
	'''
		Creates (if necessary) a folder that is namepsaced to the lot, 
		and allows reading and writing files there.  File names are not
		namespaced (because the folder is).
	'''

	@saves_args
	def __init__(self, path, dirname, *args, **kwargs):
		self.whitelist = kwargs.pop('whitelist', None)
		self.blacklist = kwargs.pop('blacklist', None)
		super(Folder, self).__init__(path, dirname, *args, **kwargs)

	def get_fname(self, fname):
		return os.path.join(self.get_path(), fname)


	def prepare_to_open(self, fname, mode):

		# check if the folder exists yet, if yes, move on
		if os.path.isdir(self.get_path()):
			pass

		# if a file (rather than folder) exists, it's an error
		elif os.path.exists(self.get_path()):
			raise IOError(
				'Folder: the given path corresponds to an existing '
				'*file*.  I cannot create a directory here: %s' 
				% self.get_path()
			)

		# if not, make the folder
		else:
			os.mkdir(self.get_path())

		# check if the specific file exists (as a folder or file)
		if os.path.isdir(self.get_fname(fname)):

			raise IOError(
				'Folder: a folder exists there. '
				'I can\'t make a file, even if clobber is True: %s' 
				% self.get_fname(fname)
			)

		# if we're opening in write mode, don't overwrite an existing file
		# unless in clobber mode
		if 'w' in mode and os.path.isfile(self.get_fname(fname)):

			if self.get_clobber():
				print '\tINFO: clobbered %s' % self.get_fname(fname)

			else:
				raise IOError(
					'Folder: a file already exists there. '
					'I do not overwrite by default: %s' % self.get_fname(fname)
				)


	def open(self, fname, mode):
		self.prepare_to_open(fname, mode)
		return open(self.get_fname(fname), mode)


	def __iter__(self):
		# make a fresh list of all the files in the dir
		files = [
			os.path.abspath(os.path.join(self.get_path(), f))
			for f in os.listdir(self.get_path()) 
		]
		self.files = [f for f in files if os.path.isfile(f)]
		return self


	def next(self):

		try:
			fname = self.files.pop()
		except IndexError:
			raise StopIteration

		if self.whitelist is not None:
			if self.whitelist.match(fname):
				return fname
			else:
				return self.next()

		if self.blacklist is not None:
			if self.blacklist.match(fname):
				return self.next()
			else:
				return fname

		return fname


	#TODO: define exists
	def exists(self):
		pass




import shutil
import unittest
from unittest import TestCase
from run import Runner, RunnerException
from task import Task, TaskException, MarkedTask, SimpleTask
from resource import Resource, File, Folder
import os


TEST_DIR = 'temporary_testing_files'

#TODO: test that the file resource creates folders along the path if needed
#	when opened in 'a' or 'w' modes

def touch(fname):
	open(fname, 'a').close()


class TestFolderResource(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)
		shutil.rmtree('./linguini_markers')

	def test_basic_write(self):

		dir_name = 'foldir'
		file_name = 'fyle.txt'
		lot_name = 'my_lot'

		class MyTask(SimpleTask):
			outputs = Folder(TEST_DIR, dir_name)
			was_run = False
			def run(self):
				self.was_run = True
				self.outputs.open(file_name, 'w').write('yo')

		class MyRunner(Runner):
			lot = lot_name
			tasks = {
				'task': MyTask()
			}

		MyRunner().run()
		expected_dir = os.path.join(TEST_DIR, '%s_%s' % (lot_name, dir_name))
		expected_file = os.path.join(expected_dir, file_name)
		self.assertTrue(os.path.isdir(expected_dir))
		self.assertTrue(os.path.isfile(expected_file))


class TestRunner(TestCase):

	def test_null_runner(self):

		# if you don't define a lot, an exception is raised
		runner = Runner()
		with self.assertRaises(RunnerException):
			runner.run()

		# A null runner can be run, once you specify the lot
		class MyRunner(Runner):
			lot = 'my_lot'

		runner = MyRunner()
		runner.run()


	def test_run_until(self):

		class MyTask(Task):
			outputs = None
			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False
			def run(self):
				self.was_run = True

		task0 = MyTask()
		task1 = MyTask()
		task2 = MyTask()
		task3 = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2': (task2, 'task1', 'task0'),
				'task3': (task3, 'task2')
			}

		my_runner = MyRunner()
		my_runner.run(until='task2')

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertFalse(task3.was_run)


	def test_run_until_classdef(self):

		class MyTask(Task):
			outputs = None
			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False
			def run(self):
				self.was_run = True

		task0 = MyTask()
		task1 = MyTask()
		task2 = MyTask()
		task3 = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			until = 'task2'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2': (task2, 'task1', 'task0'),
				'task3': (task3, 'task2')
			}

		my_runner = MyRunner()
		my_runner.run()

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertFalse(task3.was_run)


	def test_run_until_end(self):

		class MyTask(Task):
			outputs = None
			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False
			def run(self):
				self.was_run = True

		task0 = MyTask()
		task1 = MyTask()
		task2 = MyTask()
		task3 = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2': (task2, 'task1', 'task0'),
				'task3': (task3, 'task2')
			}

		my_runner = MyRunner()
		my_runner.run()

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertTrue(task3.was_run)

	
	def test_catch_cyclical_dependency(self):

		class MyTask(SimpleTask):
			def run(self):
				pass

		# a simple cycle
		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {
				'task0': (MyTask(), 'task1'),
				'task1': (MyTask(), 'task0'),
			}

		with self.assertRaises(RunnerException):
			MyRunner().run()

		# check that the cyclical dependency is reported correctly
		okay, problem = MyRunner().check_schedule()
		self.assertEqual(
			problem,
			'cyclical dependency: task0 -> task1 -> task0'
		)



		# a bit more complicated cycle
		class MoRunner(Runner):
			lot = 'mo_lot'
			tasks = {
				'task0': (MyTask(), 'task1', 'task2'),
				'task1': (MyTask(), 'task3', 'task4'),
				'task2': (MyTask(), 'task5', 'task6'),
				'task3': (MyTask(), 'task2'),
				'task4': MyTask(),
				'task5': (MyTask(), 'task1'),
				'task6': MyTask()
			}
			
		with self.assertRaises(RunnerException):
			MoRunner().run()

		# check that the cyclical dependency is reported correctly
		okay, problem = MoRunner().check_schedule()
		self.assertEqual(
			problem,
			'cyclical dependency: task1 -> task3 -> task2 -> task5 -> task1'
		)

		# here we point to a potentially previously scheduled task, but 
		# there's no cycle -- it should work
		class MiRunner(Runner):
			lot = 'mo_lot'
			tasks = {
				'task0': (MyTask(), 'task1', 'task2'),
				'task1': (MyTask(), 'task3', 'task4'),
				'task2': (MyTask(), 'task5', 'task6'),
				'task3': (MyTask(), 'task2'),
				'task4': MyTask(),
				'task5': (MyTask(), 'task6'),
				'task6': MyTask()
			}
			
		MiRunner().run()

	def test_catch_nonexistent_dependency(self):

		class MyTask(SimpleTask):
			def run(self):
				pass

		# task 1 is not defined!
		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {
				'task0': (MyTask(), 'task1'),
			}

		with self.assertRaises(RunnerException):
			MyRunner().run()

		# task7 is not defined!
		class MoRunner(Runner):
			lot = 'mo_lot'
			tasks = {
				'task0': (MyTask(), 'task1', 'task2'),
				'task1': (MyTask(), 'task3', 'task4'),
				'task2': (MyTask(), 'task5', 'task6', 'task7'),
				'task3': (MyTask(), 'task2'),
				'task4': MyTask(),
				'task5': (MyTask(), 'task6'),
				'task6': MyTask()
			}
			
		with self.assertRaises(RunnerException):
			MoRunner().run()

class TestMarkerTask(TestCase):
	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)

	def test_mark_marks(self):

		
		class MyTask(MarkedTask):
			marker_path = TEST_DIR
			outputs = None
			num_runs = 0
			def run(self):
				self.num_runs += 1

		task1 = MyTask()
		task_name = 'task1'
		lot_name = 'my_lot'

		class MyRunner(Runner):
			lot = lot_name
			tasks = {
				task_name: task1
			}

		my_runner = MyRunner()

		# run the runner
		my_runner.run()
		# check that the task was run
		self.assertTrue(task1.num_runs == 1)

		# run again
		my_runner.run()
		# this task should have been skipped due to existence of marker file
		self.assertTrue(task1.num_runs == 1)

		# remove the marker file
		os.remove(
			os.path.join(TEST_DIR, '%s_%s.marker' % (lot_name, task_name))
		)

		# run again
		my_runner.run()
		# the task should run again
		self.assertTrue(task1.num_runs == 2)


	def test_marker_as_mixin(self):
		'''
			this test is similar to the last, but it tests a situation where
			a pre-existing task is augmented using the mixin approach.
			rather than MyTask inheriting directly from MarkedTask,
			MyTask is defined on top of Task, and then an augmented version
			is made afterward, MyMarkedTask, by using MarkedTask as a mixin.

			This tests all the conditions in test_marker_marks, but it also
			tests that if MyTask overrides a method in Task, augmenting 
			with MarkedTask as a mixin preserves that.

			Of course, the `exists()` function is not preserved -- changing 
			that is the whole reason for using the MarkedTask as a mixin.
		'''
		
		class MyTask(Task):
			ready_override = False
			after_override = False
			def get_ready(self, lot, pilot, name, clobber):
				super(MyTask, self).get_ready(
					lot, pilot, name, clobber
				)
				self.ready_override = True

			def _after(self):
				super(MyTask, self)._after()
				self.after_override = True


		class MyMarkedTask(MarkedTask, MyTask):
			marker_path = TEST_DIR
			outputs = None
			num_runs = 0
			def run(self):
				self.num_runs += 1

		task1 = MyMarkedTask()
		task_name = 'task1'
		lot_name = 'my_lot'

		class MyRunner(Runner):
			lot = lot_name
			tasks = {
				task_name: task1
			}

		my_runner = MyRunner()

		# run the runner
		my_runner.run()
		# check that the task was run
		self.assertTrue(task1.num_runs == 1)

		# check that the version of get_ready from MyTask was run
		self.assertTrue(task1.ready_override)
		# check that the version of after from MyTask was run
		self.assertTrue(task1.after_override)

		# run again
		my_runner.run()
		# this task should have been skipped due to existence of marker file
		self.assertTrue(task1.num_runs == 1)

		# remove the marker file
		os.remove(
			os.path.join(TEST_DIR, '%s_%s.marker' % (lot_name, task_name))
		)

		# run again
		my_runner.run()
		# the task should run again
		self.assertTrue(task1.num_runs == 2)


class TestSimpleTask(TestCase):

	TEST_DIR = 'linguini_markers'
	def tearDown(self):
		shutil.rmtree(self.TEST_DIR)

	def test_simple_task(self):
		'''
			Simple task is based on marked task, and provides the simplest
			interface for turning some piece of functionality into a linguini
			task.  One simply defines the run function, that's all
		'''

		class MyTask(SimpleTask):
			num_runs = 0
			def run(self):
				self.num_runs += 1

		task1 = MyTask()
		task_name = 'task1'
		lot_name = 'my_lot'

		class MyRunner(Runner):
			lot = lot_name
			tasks = {
				task_name: task1
			}

		my_runner = MyRunner()

		# run the runner
		my_runner.run()
		# check that the task was run
		self.assertTrue(task1.num_runs == 1)

		# run again
		my_runner.run()
		# this task should have been skipped due to existence of marker file
		self.assertTrue(task1.num_runs == 1)

		# remove the marker file
		os.remove(
			os.path.join(self.TEST_DIR, '%s_%s.marker' % (lot_name, task_name))
		)

		# run again
		my_runner.run()
		# the task should run again
		self.assertTrue(task1.num_runs == 2)



class TestCompoundRunner(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)

	def test_run_compound(self):
		'''
			Tests that a compound runner does run the entire thing
		'''
		
		class MyTask(Task):

			def _outputs(self):
				fname = 'test%d.txt' % self.parameters['file_num']
				return File(TEST_DIR, fname)

			was_run = False		# Track whether the task was run
			def exists(self):	# ensures the task will not be skipped
				return False

			def run(self):
				fh = self.outputs.open('w')
				fh.write('yo')
				fh.close()
				self.was_run = True

		task0 = MyTask(file_num=0)
		task1 = MyTask(file_num=1)
		task2 = MyTask(file_num=2)
		task3 = MyTask(file_num=3)
		task4 = MyTask(file_num=4)
		task5 = MyTask(file_num=5)
		task6 = MyTask(file_num=6)
		task7 = MyTask(file_num=7)

		class MyRunner(Runner):
			outputs = None
			lot = 'a'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2': (task2, 'task1', 'task0'),
				'task3': (task3, 'task2')
			}
			def exists(self):
				return False

		class MoRunner(Runner):
			outputs = None
			lot = 'b'
			tasks = {
				'task0':task4,
				'task1':task5,
				'task2': (task6, 'task1', 'task0'),
				'task3': (task7, 'task2')
			}
			def exists(self):
				return False

		class MasterRunner(Runner):
			lot = '1'
			tasks = {
				'run0': MyRunner(),
				'run1': (MoRunner(), 'run0'),
			}

		my_runner = MasterRunner()
		my_runner.run()

		self.assertTrue(task0.was_run)
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertTrue(task3.was_run)
		self.assertTrue(task4.was_run)
		self.assertTrue(task5.was_run)
		self.assertTrue(task6.was_run)
		self.assertTrue(task7.was_run)

		# the lot of master runner is inherited by both sub-runners
		expected_files = ['1_test%d.txt' % d for d in range(8)]
		self.assertItemsEqual(expected_files, os.listdir(TEST_DIR))


	def test_run_compound_partially_complete(self):
		'''
			Tests that a compound runner does runs the entire but checks
			whether the components have been run before
		'''
		
		class MyTask(Task):

			def _outputs(self):
				fname = 'test%d.txt' % self.parameters['file_num']
				return File(TEST_DIR, fname)

			was_run = False		# Track whether the task was run

			def run(self):
				fh = self.outputs.open('w')
				fh.write('yo')
				fh.close()
				self.was_run = True

		task0 = MyTask(file_num=0)
		task1 = MyTask(file_num=1)
		task2 = MyTask(file_num=2)
		task3 = MyTask(file_num=3)
		task4 = MyTask(file_num=4)
		task5 = MyTask(file_num=5)
		task6 = MyTask(file_num=6)
		task7 = MyTask(file_num=7)

		class MyRunner(Runner):
			outputs = None
			lot = 'a'
			tasks = {
				'task0':task0,
				'task1':task1,
				'task2': (task2, 'task1', 'task0'),
				'task3': (task3, 'task2')
			}
			def exists(self):
				return False

		class MoRunner(Runner):
			outputs = None
			lot = 'b'
			tasks = {
				'task0':task4,
				'task1':task5,
				'task2': (task6, 'task1', 'task0'),
				'task3': (task7, 'task2')
			}
			def exists(self):
				return False

		my_runner = MyRunner()
		mo_runner = MoRunner()


		class MasterRunner(Runner):
			lot = '1'
			tasks = {
				'run0': my_runner,
				'run1': (mo_runner, 'run0')
			}

		# make some resources so they won't be run
		touch(os.path.join(TEST_DIR, '1_test0.txt'))
		touch(os.path.join(TEST_DIR, '1_test6.txt'))

		my_runner = MasterRunner()
		my_runner.run()

		# these tasks should not have been run
		self.assertFalse(task0.was_run)
		self.assertFalse(task6.was_run)

		# these tasks should have been run
		self.assertTrue(task1.was_run)
		self.assertTrue(task2.was_run)
		self.assertTrue(task3.was_run)
		self.assertTrue(task4.was_run)
		self.assertTrue(task5.was_run)
		self.assertTrue(task7.was_run)

		# all the files should exist at this point
		expected_files = ['1_test%d.txt' % d for d in range(8)]

		# note, b_test4.txt and b_test5.txt never get made because
		# they aren't needed.

		self.assertItemsEqual(expected_files, os.listdir(TEST_DIR))





class TestResource(TestCase):
	

	def test_exists_not_implemented(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			pass

		class MyTask(Task):
			outputs = MyResource()

			def run(self):
				print 'hello world'

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}

		runner = MyRunner()
		with self.assertRaises(NotImplementedError):
			runner.run()


	def test_exists_implemented(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			def exists(self):
				return True

		class MyTask(Task):
			outputs = MyResource()

			def run(self):
				print 'hello world'

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}

		runner = MyRunner()
		runner.run()



class TestTask(TestCase):

	def test_no_output(self):	

		# a task has to have some outputs
		class MyTask(Task):
			def run(self):
				print 'hello world'

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}

		runner = MyRunner()

		with self.assertRaises(TaskException):
			runner.run()


	def test_run_not_implemented(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = Resource()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':MyTask()}

		runner = MyRunner()

		with self.assertRaises(NotImplementedError):
			runner.run()


	def test_simple_task_gets_run(self):
		'''
			Test that a task that doesn't define a run function raises a 
			not implemented error
		'''
		
		class MyResource(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = MyResource()
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}

		runner = MyRunner()
		runner.run()

		self.assertTrue(my_task.was_run)


	def test_simple_task_not_run(self):
		'''
			Test that a task whose only resource already exists does not get 
			run.
		'''
		
		class MyResource(Resource):
			def exists(self):
				return True

		class MyTask(Task):
			outputs = MyResource()
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}

		runner = MyRunner()
		runner.run()

		self.assertFalse(my_task.was_run)



	def test_simple_task_many_resources_dict_run(self):
		'''
			Test that a task with many resources as a list,
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True

		class Noexists(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = {
				'one': Exists(),
				'two': Noexists(),
				'three': Exists(),
				'four': Exists()
			}
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}

		runner = MyRunner()
		runner.run()

		self.assertTrue(my_task.was_run)


	def test_simple_task_many_resources_dict_no_run(self):
		'''
			Test that a task with many resources as a list, not 
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True

		class MyTask(Task):
			outputs = {
				'one': Exists(),
				'two': Exists(),
				'three': Exists(),
				'four': Exists()
			}
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}

		runner = MyRunner()
		runner.run()

		self.assertFalse(my_task.was_run)


	def test_simple_task_many_resources_no_run(self):
		'''
			Test that a task with many resources as a list, not 
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True


		class MyTask(Task):
			outputs = (Exists(), Exists(), Exists(), Exists())
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}

		runner = MyRunner()
		runner.run()

		self.assertFalse(my_task.was_run)




	def test_simple_task_many_resources(self):
		'''
			Test that a task with many resources as a list, not 
			gets run if at least one of them does not exist
		'''
		
		class Exists(Resource):
			def exists(self):
				return True

		class Noexists(Resource):
			def exists(self):
				return False

		class MyTask(Task):
			outputs = (Exists(), Exists(), Noexists(), Exists())
			was_run = False

			def run(self):
				self.was_run = True

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_lot'
			tasks = {'my_task':my_task}

		runner = MyRunner()
		runner.run()

		self.assertTrue(my_task.was_run)


class TestStatic(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)
		if os.path.exists('./linguini_markers'):
			shutil.rmtree('./linguini_markers')


	def test_ignore_pilot_resource(self):
		'''
			Tests that, when a resource is invoked with the keyword argument
			ignore_pilot=True, it will not put 'pilot_' in front of its 
			file names.
		'''

		LOT_NAME = 'my_lot'

		# make a files to be read, one prepends pilot, the other doesn't
		open(os.path.join(TEST_DIR, '%s_in_A.txt'%LOT_NAME), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTask(SimpleTask):
			outputs = {
				'A': File(TEST_DIR, 'out_A.txt', ignore_pilot=True),
				'B': File(TEST_DIR, 'out_B.txt'),
			}
			inputs = {
				'A': File(TEST_DIR, 'in_A.txt', ignore_pilot=True),
				'B': File(TEST_DIR, 'in_B.txt')
			}

			def run(self):
				self.inputs['A'].open('r').read()
				self.inputs['B'].open('r').read()
				self.outputs['A'].open('w').write('yo')
				self.outputs['B'].open('w').write('yo')

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {'my_task':MyTask()}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_out_A.txt'%LOT_NAME)))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))


	def test_share_resource(self):
		'''
			Tests that, when a resource is invoked with the keyword argument
			share=True, it will not prepend the lot name in_front of its
			file names.
		'''

		LOT_NAME = 'my_lot'

		# make a files to be read, one prepends pilot, the other doesn't
		open(os.path.join(TEST_DIR, 'in_A.txt'), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTask(SimpleTask):
			outputs = {
				'A': File(TEST_DIR, 'out_A.txt', share=True),
				'B': File(TEST_DIR, 'out_B.txt'),
			}
			inputs = {
				'A': File(TEST_DIR, 'in_A.txt', share=True),
				'B': File(TEST_DIR, 'in_B.txt')
			}

			def run(self):
				self.inputs['A'].open('r').read()
				self.inputs['B'].open('r').read()
				self.outputs['A'].open('w').write('yo')
				self.outputs['B'].open('w').write('yo')

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {'my_task':MyTask()}

		# this will cause an error if the input prepended the input file name
		MyRunner().run()

		# check that the output was also not prepended by pilot
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, 'out_A.txt')))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_out_B.txt'%LOT_NAME)))


	def test_static_resource(self):
		'''
			Tests that, when a resource is invoked with the keyword argument
			static=True, it will not put the lot name nor 'pilot_' in front 
			of its file names.
		'''

		LOT_NAME = 'my_lot'

		# make a files to be read, one prepends pilot, the other doesn't
		open(os.path.join(TEST_DIR, 'in_A.txt'), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTask(SimpleTask):
			outputs = {
				'A': File(TEST_DIR, 'out_A.txt', static=True),
				'B': File(TEST_DIR, 'out_B.txt'),
			}
			inputs = {
				'A': File(TEST_DIR, 'in_A.txt', static=True),
				'B': File(TEST_DIR, 'in_B.txt')
			}

			def run(self):
				self.inputs['A'].open('r').read()
				self.inputs['B'].open('r').read()
				self.outputs['A'].open('w').write('yo')
				self.outputs['B'].open('w').write('yo')

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {'my_task':MyTask()}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, 'out_A.txt')))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))


class TestStaticTask(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)
		if os.path.exists('./linguini_markers'):
			shutil.rmtree('./linguini_markers')


	def test_ignore_pilot_task(self):
		'''
			Tests that, when a task is invoked with the keyword argument
			ignore_pilot=True, it will invoke it's resources with pilot=False.
			The resources will therefore not have their filenames prepended by
			'pilot_'.
		'''

		LOT_NAME = 'my_lot'

		# make a files to be read, one prepends pilot, the other doesn't
		open(os.path.join(TEST_DIR, '%s_in_A.txt'%LOT_NAME), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTaskA(SimpleTask):
			outputs = File(TEST_DIR, 'out_A.txt')
			inputs = File(TEST_DIR, 'in_A.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')


		class MyTaskB(SimpleTask):
			outputs = File(TEST_DIR, 'out_B.txt')
			inputs = File(TEST_DIR, 'in_B.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')


		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'my_static_task':MyTaskA(ignore_pilot=True),
				'my_task':MyTaskB()
			}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_out_A.txt'%LOT_NAME)))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))


	def test_share_task(self):
		'''
			Tests that, when a task is invoked with the keyword argument
			share=True, it will invoke it's resources with lot=None.  The 
			resources will therefore not have their filenames prepended by
			the lot.
		'''

		LOT_NAME = 'my_lot'

		# make a files to be read, one prepends pilot, the other doesn't
		open(os.path.join(TEST_DIR, 'pilot_in_A.txt'), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTaskA(SimpleTask):
			outputs = File(TEST_DIR, 'out_A.txt')
			inputs = File(TEST_DIR, 'in_A.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')


		class MyTaskB(SimpleTask):
			outputs = File(TEST_DIR, 'out_B.txt')
			inputs = File(TEST_DIR, 'in_B.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')


		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'my_static_task':MyTaskA(share=True),
				'my_task':MyTaskB()
			}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, 'pilot_out_A.txt')))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))


	def test_static_task(self):
		'''
			Tests that, when a task is invoked with the keyword argument
			static=True, it will invoke it's resources with lot=None and
			pilot=False.  The resources will therefore not have their
			filenames prepended.
		'''

		LOT_NAME = 'my_lot'

		# make a files to be read, one prepends pilot, the other doesn't
		open(os.path.join(TEST_DIR, 'in_A.txt'), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTaskA(SimpleTask):
			outputs = File(TEST_DIR, 'out_A.txt')
			inputs = File(TEST_DIR, 'in_A.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')


		class MyTaskB(SimpleTask):
			outputs = File(TEST_DIR, 'out_B.txt')
			inputs = File(TEST_DIR, 'in_B.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')


		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'my_static_task':MyTaskA(static=True),
				'my_task':MyTaskB()
			}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, 'out_A.txt')))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))


class TestStaticRunner(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)
		if os.path.exists('./linguini_markers'):
			shutil.rmtree('./linguini_markers')

	def test_static_runner(self):
		'''
			Tests that, when a runner is invoked with the keyword argument
			static=True, it will not inherit it's parent's lot name, and it
			will always run with pilot=False, even if it's parent is run
			with pilot=True.
		'''

		LOT_NAME = 'my_lot'
		LOT_NAME_A = 'my_lot_A'
		LOT_NAME_B = 'my_lot_B'

		# make a files to be read, one prepends pilot, the other doesn't
		# and one uses the lot name of the subrunner, the other uses that
		# of the parent runner
		open(os.path.join(
			TEST_DIR, '%s_in_A.txt' % LOT_NAME_A), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTaskA(SimpleTask):
			outputs = File(TEST_DIR, 'out_A.txt')
			inputs = File(TEST_DIR, 'in_A.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')

		class MyTaskB(SimpleTask):
			outputs = File(TEST_DIR, 'out_B.txt')
			inputs = File(TEST_DIR, 'in_B.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')

		class SubRunnerA(Runner):
			lot = LOT_NAME_A
			tasks = {'my_task':MyTaskA()}

		class SubRunnerB(Runner):
			lot = LOT_NAME_B
			tasks = {'my_task':MyTaskB()}

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'runner_A': SubRunnerA(static=True),
				'runner_B': SubRunnerB()
			}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_out_A.txt'%LOT_NAME_A)))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))


	def test_ignore_pilot_runner(self):
		'''
			Tests that, when a runner is instantiated with the keyword 
			argument ignore_pilot=True, it will be run with pilot=False even
			if its parent is run with pilot=True.
		'''

		LOT_NAME = 'my_lot'
		LOT_NAME_A = 'my_lot_A'
		LOT_NAME_B = 'my_lot_B'

		# make a files to be read, one prepends pilot, the other doesn't
		open(os.path.join(
			TEST_DIR, '%s_in_A.txt' % LOT_NAME), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTaskA(SimpleTask):
			outputs = File(TEST_DIR, 'out_A.txt')
			inputs = File(TEST_DIR, 'in_A.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')

		class MyTaskB(SimpleTask):
			outputs = File(TEST_DIR, 'out_B.txt')
			inputs = File(TEST_DIR, 'in_B.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')

		class SubRunnerA(Runner):
			lot = LOT_NAME_A
			tasks = {'my_task':MyTaskA()}

		class SubRunnerB(Runner):
			lot = LOT_NAME_B
			tasks = {'my_task':MyTaskB()}

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'runner_A': SubRunnerA(ignore_pilot=True),
				'runner_B': SubRunnerB()
			}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		# but both sub-runners still inherit their parent's lot name 
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_out_A.txt'%LOT_NAME)))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))



	def test_share_runner(self):
		'''
			Tests that, when a runner is instantiated with the keyword 
			argument ignore_pilot=True, it will be run with pilot=False even
			if its parent is run with pilot=True.
		'''

		LOT_NAME = 'my_lot'
		LOT_NAME_A = 'my_lot_A'
		LOT_NAME_B = 'my_lot_B'

		# make files to be read, one uses lot name of the subrunner, the other
		# uses lot name of the parent runner.  Both prepend with pilot.
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_A.txt' % LOT_NAME_A), 'w').write('oy')
		open(os.path.join(
			TEST_DIR, '%s_pilot_in_B.txt'%LOT_NAME), 'w').write('oy')

		class MyTaskA(SimpleTask):
			outputs = File(TEST_DIR, 'out_A.txt')
			inputs = File(TEST_DIR, 'in_A.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')

		class MyTaskB(SimpleTask):
			outputs = File(TEST_DIR, 'out_B.txt')
			inputs = File(TEST_DIR, 'in_B.txt')

			def run(self):
				self.inputs.open('r').read()
				self.outputs.open('w').write('yo')

		class SubRunnerA(Runner):
			lot = LOT_NAME_A
			tasks = {'my_task':MyTaskA()}

		class SubRunnerB(Runner):
			lot = LOT_NAME_B
			tasks = {'my_task':MyTaskB()}

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'runner_A': SubRunnerA(share=True),
				'runner_B': SubRunnerB()
			}

		# this will cause an error if the input prepended the input file name
		MyRunner().run(pilot=True)

		# check that the output was also not prepended by pilot
		# but both sub-runners still inherit their parent's lot name 
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_A.txt'%LOT_NAME_A)))
		self.assertTrue(os.path.isfile(
			os.path.join(TEST_DIR, '%s_pilot_out_B.txt'%LOT_NAME)))


class TestClobberInheritance(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)

	def tearDown(self):
		shutil.rmtree(TEST_DIR)
		if os.path.exists('./linguini_markers'):
			shutil.rmtree('./linguini_markers')

	
	def test_resource_inheritance(self):

		# a Resource's clobber flag is set by get_ready
		my_file = File('.', 'file.txt')
		my_file.get_ready(lot='my_lot', pilot=False, name='name', 
			clobber=True)
		self.assertTrue(my_file.get_clobber())

		# it's overriden by instantiation keyword argument
		self.assertTrue(File('.', 'file.txt', clobber=True).get_clobber())

		# instantiation gets the last word
		my_file = File('.', 'file.txt', clobber=True)
		my_file.get_ready(lot='my_lot', pilot=False, name='name', 
			clobber=False)
		self.assertTrue(my_file.get_clobber())

		# instantiation gets the last word
		my_file = File('.', 'file.txt', clobber=False)
		my_file.get_ready(lot='my_lot', pilot=False, name='name', 
			clobber=True)
		self.assertFalse(my_file.get_clobber())


		class MyFile(File):
			clobber = True

		# this resource clobbers by default
		self.assertTrue(MyFile('.', 'file.txt').get_clobber())

		# that behavior is overriden by instantiation keyword argument
		self.assertFalse(MyFile('.', 'file.txt', clobber=False).get_clobber())

		# it's not overridden by argument passed to get_ready
		my_file = MyFile('.', 'file.txt')
		my_file.get_ready(lot='my_lot', pilot=False, name='name', 
			clobber=False)
		self.assertTrue(my_file.get_clobber())


		class MyFile(File):
			clobber = False

		# this resource does not clobber by default
		self.assertFalse(MyFile('.', 'file.txt').get_clobber())

		# that behavior is overriden by instantiation keyword argument
		self.assertTrue(MyFile('.', 'file.txt', clobber=True).get_clobber())

		# it's not overridden by argument passed to get_ready
		my_file = MyFile('.', 'file.txt')
		my_file.get_ready(lot='my_lot', pilot=False, name='name', 
			clobber=True)
		self.assertFalse(my_file.get_clobber())



	def test_task_inheritance(self):

		LOT_NAME = 'my_lot'

		open(os.path.join(TEST_DIR, '%s_a.txt'%LOT_NAME), 'w').write('no')
		open(os.path.join(TEST_DIR, '%s_b.txt'%LOT_NAME), 'w').write('no')

		class MyTask(SimpleTask):
			outputs = {
				'A': File(TEST_DIR, 'a.txt'),
				'B': File(TEST_DIR, 'b.txt')
			}
			def run(self):
				self.outputs['A'].open('w').write('yes')
				self.outputs['B'].open('w').write('yes')

		task = MyTask()

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'task1' : task
			}


		# by default task has clobber = False 
		with self.assertRaises(IOError):
			MyRunner().run()

		self.assertFalse(task.get_clobber())
		self.assertEqual(
			'no', 
			open(os.path.join(TEST_DIR, '%s_a.txt'%LOT_NAME), 'r').read()
		)
		self.assertEqual(
			'no', 
			open(os.path.join(TEST_DIR, '%s_b.txt'%LOT_NAME), 'r').read()
		)

		# contents should not have changed, but reset for certainty
		open(os.path.join(TEST_DIR, '%s_a.txt'%LOT_NAME), 'w').write('no')
		open(os.path.join(TEST_DIR, '%s_b.txt'%LOT_NAME), 'w').write('no')

		# this gets overidden by keyword argument to runner
		MyRunner().run(clobber=True)

		self.assertTrue(task.get_clobber())
		self.assertEqual(
			'yes', 
			open(os.path.join(TEST_DIR, '%s_a.txt'%LOT_NAME), 'r').read()
		)
		self.assertEqual(
			'yes', 
			open(os.path.join(TEST_DIR, '%s_b.txt'%LOT_NAME), 'r').read()
		)


		task = MyTask(clobber=True)

		class MyRunner(Runner):
			lot = LOT_NAME
			tasks = {
				'task1' : task
			}



class TestClobberMode(TestCase):

	def setUp(self):
		os.mkdir(TEST_DIR)


	def tearDown(self):
		shutil.rmtree(TEST_DIR)
		if os.path.exists('./linguini_markers'):
			shutil.rmtree('./linguini_markers')


	def test_folder_resource(self):
		'''
			test that when a task is run in clobber mode, the Folder
			resource is willing to overwrite files
		'''
		subfolder = os.path.join(TEST_DIR, 'subfolder')
		os.mkdir(subfolder)

		class MyTask(SimpleTask):
			outputs = Folder(TEST_DIR, 'subfolder')
			def run(self):
				self.outputs.open('yo.txt', 'w').write('yo')

		class MyRunner(Runner):
			lot='my_lot'
			tasks = {
				'task': MyTask()
			}

		MyRunner().run(clobber=True)
		MyRunner().run(clobber=True)

	
	def test_simple_task_inherit(self):

		runs = []
		class MyTask(SimpleTask):
			def run(self):
				runs.append('yo')

		class MyRunner(Runner):
			lot = 'my_runner'
			tasks = {
				'task': MyTask()
			}

		# to begin with we haven't run the runner yet
		self.assertEqual(len(runs), 0)

		# after first call to run, we can see it has run once
		MyRunner().run()
		self.assertEqual(len(runs), 1)

		# second call to run does not cause it to run 
		MyRunner().run()
		self.assertEqual(len(runs), 1)

		# we can force it to run again using clobber
		MyRunner().run(clobber=True)
		self.assertEqual(len(runs), 2)

		# now we test the clobber argument when added to the Tasks class
		# definition
		

	def test_simple_task_invocation(self):

		runs = []

		class MyTask(SimpleTask):
			clobber = True
			def run(self):
				runs.append('yo')

		my_task = MyTask()

		class MyRunner(Runner):
			lot = 'my_runner'
			tasks = {
				'task': my_task
			}


		# to begin with we haven't run the runner yet
		self.assertEqual(len(runs), 0)

		# after first call to run, we can see it has run once
		MyRunner().run()
		self.assertEqual(len(runs), 1)

		self.assertTrue(my_task.get_clobber())

		# second call will run it, because the class def has clobber=True
		MyRunner().run()
		self.assertEqual(len(runs), 2)

		## we can force it to run again using clobber
		#MyRunner().run(clobber=True)
		#self.assertEqual(len(runs), 2)



		





if __name__ == '__main__':
	unittest.main()




from task import Task
from resource import Resource, File

def as_list(item):
	if isinstance(item, dict):
		return item.values()

	if hasattr(item, '__iter__'):
		return list(item)

	return [item]


class RunnerException(Exception):
	pass


class Runner(Task):

	# The null runner has no tasks, but is valid
	tasks = {}
	until = None
	lot = None
	outputs = None
	inputs = None

	def _tasks(self):
		return self.tasks


	def _layout(self):
		return self.layout


	def _until(self):
		return self.until


	def get_all_outputs(self):
		'''
			Collect all the outputs from the tasks that end this runner.
			These are taken to be the outputs of the runner itself.
		'''
		return reduce(
			lambda x,y: x+self.get_task(y).get_all_outputs(),
			self.tasks.keys(),
			[]
		)

	def get_task(self, task_name):
		return as_list(self.tasks[task_name])[0]

	def get_lot(self):

		# if self.share is True, use the lot from the class definition
		if self.share or self.static:
			if self.lot is None:
				raise RunnerException(
					'lot is not defined in %s' % self.__class__.__name__)
			return self.lot

		# normally, inherit lot
		if self.inherited_lot is not None:
			return self.inherited_lot

		# otherwise instantiated lot takes precedence
		if self.instance_lot is not None:
			return self.instance_lot

		# lastly, use the lot in the class definition
		if self.lot is None:
			raise RunnerException(
				'lot is not defined in %s' % self.__class__.__name__)
		return self.lot


	def get_ready(
			self,
			lot=None,
			pilot=False,
			name='main',
			clobber=False
		):

		if self.is_ready():
			return

		self.inherited_lot = lot
		self.inherited_pilot = pilot
		self.inherited_clobber = clobber

		self.name = name

		# at this point we should have a valid lot
		if not isinstance(self.get_lot(), basestring):
			raise RunnerException(
				'lot must be string-like (in %s).' % self.__class__.__name__)

		# resolve tasks
		self.tasks = self._tasks()

		# get all the tasks ready
		for task_name in self.tasks:

			task = as_list(self.tasks[task_name])[0]

			task.get_ready(
				lot=self.get_lot(), pilot=self.get_pilot(), name=task_name, 
				clobber=self.get_clobber()
			)

		# it isn't necessary to specify a run's outputs, because they are
		# taken to be those of the END tasks
		self.outputs = self.get_all_outputs()

		# mark as ready
		self._is_ready = True


	def recursively_schedule(self, task_names):
		'''
			accepts the name of a task, or an iterable thereof, and schedules
			those among them which are not done, as well as their 
			not-done-dependencies.  Avoids double-scheduling tasks.

			if self.clobber is True, then it schedules tasks even if already 
			done.
		'''

		task_names = as_list(task_names)
		scheduled = set()

		for task_name in reversed(task_names):

			task_def = as_list(self.tasks[task_name])
			task = task_def[0]
			dependencies = task_def[1:]

			if not task.exists() or task.get_clobber():

				# add tasks to the schedule, without permitting duplicates
				scheduled.add(task_name)

				# schedule dependant jobs, if any
				try:
					scheduled |= (self.recursively_schedule(dependencies))
				except KeyError:
					pass
				
		return scheduled


	def check_schedule(self):

		checked_tasks = set()
		okay, problem = True, None
		for task_name in self.tasks:
			okay, problem = self.check_task(task_name, checked_tasks, [])
			if not okay:
				return okay, problem

		return okay, problem


	def check_task(self, task_name, checked_tasks, pending_tasks):

		okay = True
		pending_tasks.append(task_name)

		try:
			dependencies = as_list(self.tasks[task_name])[1:]
		except:
			return (
				False, 
				'%s is listed as a dependency but never defined in %s tasks'
				% (task_name, self.__class__.__name__)
			)

		# detect cyclical dependencies arising at this step
		cyclical_dependencies = [d for d in dependencies if d in pending_tasks]
		if len(cyclical_dependencies):
			cyc_d = cyclical_dependencies[0]
			entry = pending_tasks.index(cyc_d)
			cycle = pending_tasks[entry:] + [cyc_d]
			problem = 'cyclical dependency: %s' % ' -> '.join(cycle)
			return False, problem

		okay, problem = True, None
		for d in dependencies:

			# we don't need to re-check previously checked tasks
			if d in checked_tasks:
				continue

			# recurse, check this tasks dependencies
			okay, problem = self.check_task(d, checked_tasks, pending_tasks)
			if not okay:
				return okay, problem

		this_task_name = pending_tasks.pop()
		assert(this_task_name == task_name)
		checked_tasks.add(task_name)

		return okay, problem


	def run_schedule(self):

		# keep looping as long as their are incomplete tasks
		while len(self.schedule)>0:

			remaining = [t for t in self.schedule]

			# look at each task left in the schedule
			for task_name in remaining:

				# get the task definition
				task_def = as_list(self.tasks[task_name])
				task = task_def[0]
				dependencies = task_def[1:]

				# if the task has scheduled dependencies, we can't run it yet
				if any([d in self.schedule for d in dependencies]):
					continue

				# run the task, and remove it from the schedule
				task._run()
				self.schedule.remove(task_name)


	def _run(
			self, 
			lot=None,
			pilot=False,
			until=None,
			clobber=None,
			share=False
		):

		self.run(lot, pilot=pilot, until=until, clobber=clobber)


	def run(
			self, 
			lot=None,
			pilot=False,
			until=None,
			clobber=False,
			share=False
		):

		self.share = share

		self.get_ready(
			lot=lot, 
			pilot=pilot,
			name='main',
			clobber=clobber
		)

		# resolve until
		if until is not None:
			self.until = until

		elif not hasattr(self, 'until'):
			self.until = None

		elif hasattr(self, '_until'):
			self.until = self._until()

		#print 'Starting runner for lot %s.' % self.get_lot()

		# by default, run whatever is in 'END'
		until = until or self.until
		if until is None:
			until = self.tasks.keys()

		# ensure that the schedule can actually complete
		okay, problem = self.check_schedule()
		if not okay:
			raise RunnerException(problem)

		# schedule the necessary tasks
		self.schedule = self.recursively_schedule(until)

		# run the tasks
		self.run_schedule()





#!/usr/bin/env python
from setuptools import setup
#from distutils.core import setup

setup(
	name='linguini',
	version='1.1.6',
	description='beats both spaghetti and ravioli',
	author='edward newell',
	author_email='edward.newell@gmail.com',
	packages=['linguini'],
	scripts=['bin/linguini'],
	license='MIT',
	classifiers=[
		'Development Status :: 2 - Pre-Alpha',
		'Intended Audience :: Science/Research',
		'Intended Audience :: System Administrators',
		'Intended Audience :: Information Technology',
		'Programming Language :: Python :: 2.7',
		'Natural Language :: English',
		'Topic :: Utilities',
	],
	keywords='pipe dream'
)




import os
import sys


if len(sys.argv) == 1:
	project_name = ''
elif len(sys.argv) == 2:
	project_name = sys.argv[1] 
elif len(sys.argv) == 3:
	project_name = sys.argv[2]

settings_base = '''import os

PROJ_DIR = '%s'
DATA_DIR = os.path.join(PROJ_DIR, 'data')''' % os.path.join(
	os.getcwd(), project_name
)

runner_base = '''import sys
sys.path.append('.')
from tasks.my_first_task import MyFirstTask
from tasks.my_second_task import MySecondTask
from linguini import Runner

class MyRunner(Runner):
	lot = 'my_lot'
	tasks = {
		'task1': MyFirstTask(),
		'task2': (MySecondTask(), 'task1')
	}

if __name__=='__main__':
	MyRunner().run()
'''

first_task_base = '''from linguini import Task, SimpleTask, File, Folder
from local_settings import DATA_DIR

class MyFirstTask(SimpleTask):
	inputs = File(DATA_DIR, '0.txt', static=True)
	outputs = File(DATA_DIR, '1.txt')

	def run(self):
		contents = self.inputs.open('r').read()
		self.outputs.open('w').write(contents[::-1]) # reverses contents'''

second_task_base = '''from linguini import Task, SimpleTask, File, Folder
from local_settings import DATA_DIR

class MySecondTask(SimpleTask):
	inputs = File(DATA_DIR, '1.txt')
	outputs = File(DATA_DIR, '2.txt')

	def run(self):
		contents = self.inputs.open('r').read()
		self.outputs.open('w').write(
			'\\n'.join(['\\t'*i + c for i,c in enumerate(contents)])	# stagger
		)'''

input_base = '1234567'

DIRS = {
	'src': ['tasks', 'resources'],
	'docs': ['documentation', 'publications'],
	'data': None
}


FILES = {
	'': ('README.md', ''),
	'src': [
		('local_settings.py', settings_base),
		('run.py', runner_base)
	],
	'src/resources': ('__init__.py', ''),
	'src/tasks': [
		('__init__.py', ''),
		('my_first_task.py', first_task_base),
		('my_second_task.py', second_task_base),
	],
	'data': ('0.txt', input_base)
}




runner_base = '''import sys
sys.path.append('.')
from tasks.my_first_task import MyFirstTask
from tasks.my_second_task import MySecondTask
from linguini import Runner

class MyRunner(Runner):
	lot = 'my_lot'
	tasks = {
		'task1': MyFirstTask(),
		'task2': (MySecondTask(), 'task1')
	}

if __name__=='__main__':
	MyRunner().run()
'''


first_task_base = '''from linguini import Task, SimpleTask, File, Folder

class MyFirstTask(SimpleTask):
	inputs = File('.', '0.txt', static=True)
	outputs = File('.', '1.txt')

	def run(self):
		contents = self.inputs.open('r').read()
		self.outputs.open('w').write(contents[::-1]) # reverses contents'''


second_task_base = '''from linguini import Task, SimpleTask, File, Folder

class MySecondTask(SimpleTask):
	inputs = File('.', '1.txt')
	outputs = File('.', '2.txt')

	def run(self):
		contents = self.inputs.open('r').read()
		self.outputs.open('w').write(
			'\\n'.join(['\\t'*i + c for i,c in enumerate(contents)])
		)'''


input_base = '1234567'

DIRS = ['tasks', 'resources']

FILES = {
	'': [
		('run.py', runner_base),
		('0.txt', input_base)
	],

	'tasks': [
		('__init__.py', ''),
		('my_first_task.py', first_task_base),
		('my_second_task.py', second_task_base),
	]
}
		

